# This file is a part of the AnyBlok project
#
#    Copyright (C) 2016 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Pierre Verkest <pverkest@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import TestCase
from anyblok.registry import Registry, RegistryManager
from anyblok.blok import BlokManager
from anyblok.config import Configuration
from anyblok.environment import EnvironmentManager
from anyblok.column import Integer as Int, String as Str
from anyblok.migration import MigrationException
from anyblok.relationship import Many2Many
from contextlib import contextmanager
from sqlalchemy import Column, Integer, TEXT
from anyblok import Declarations
from sqlalchemy.exc import InternalError, IntegrityError
from unittest import skipIf
import alembic
from copy import deepcopy
from sqlalchemy.orm import clear_mappers

skipAlembicVersion = (
    tuple(int(x) for x in alembic.__version__.split('.')) < (0, 9, 6)
)


class TestMigration(TestCase):

    @classmethod
    def setUpClass(cls):
        super(TestMigration, cls).setUpClass()
        cls.init_configuration_manager()
        cls.createdb()
        BlokManager.load()

        register = Declarations.register
        Model = Declarations.Model

        cls.loaded_bloks = deepcopy(RegistryManager.loaded_bloks)
        EnvironmentManager.set('current_blok', 'anyblok-core')

        @register(Model)
        class Test:
            integer = Int(primary_key=True)
            other = Str()

        @register(Model)
        class TestUnique:
            integer = Int(primary_key=True)
            other = Str(unique=True)

        @register(Model)
        class TestFKTarget:
            integer = Int(primary_key=True)

        @register(Model)
        class TestFK:
            integer = Int(primary_key=True)
            other = Int(foreign_key=Model.TestFKTarget.use('integer'))

        @register(Model)
        class TestM2M1:
            idmodel1 = Int(primary_key=True)

        @register(Model)
        class TestM2M2:
            idmodel2 = Int(primary_key=True)
            rel_m2m = Many2Many(label="Rel", model=Model.TestM2M1,
                                join_table='reltable',
                                remote_columns='idmodel1',
                                m2m_remote_columns='idmodel1',
                                local_columns='idmodel2',
                                m2m_local_columns='idmodel2',
                                many2many='rel_m2m_inv')

        EnvironmentManager.set('current_blok', None)

    def setUp(self):
        super(TestMigration, self).setUp()
        self.registry = Registry(Configuration.get('db_name'), unittest=True)

    @classmethod
    def tearDownClass(cls):
        super(TestMigration, cls).tearDownClass()
        BlokManager.unload()
        RegistryManager.loaded_bloks = cls.loaded_bloks
        cls.dropdb()

    def tearDown(self):
        super(TestMigration, self).tearDown()
        for table in ('test', 'test2', 'othername', 'testfk', 'testfktarget',
                      'testunique', 'reltab', 'testm2m1', 'testm2m2'):
            try:
                self.registry.migration.table(table).drop()
            except:
                pass

        self.registry.migration.conn.close()
        clear_mappers()
        self.registry.close()

    @contextmanager
    def cnx(self):
        cnx = self.registry.migration.conn
        try:
            yield cnx
        except Exception:
            cnx.execute("rollback")
            raise

    def test_add_table(self):
        self.registry.migration.table().add('test2')
        self.registry.migration.table('test2')

    def fill_test_table(self):
        Test = self.registry.Test
        vals = [{'other': 'test %d' % x} for x in range(10)]
        Test.multi_insert(*vals)

    def test_add_column(self):
        t = self.registry.migration.table('test')
        t.column().add(Column('new_column', Integer))
        t.column('new_column')

    def test_add_column_in_filled_table(self):
        self.fill_test_table()
        t = self.registry.migration.table('test')
        t.column().add(Column('new_column', Integer))
        t.column('new_column')

    def test_add_not_null_column(self):
        t = self.registry.migration.table('test')
        t.column().add(Column('new_column', Integer, nullable=False))
        t.column('new_column')

    def test_add_not_null_column_in_filled_table(self):
        self.fill_test_table()
        t = self.registry.migration.table('test')
        t.column().add(Column('new_column', Integer, nullable=False))
        t.column('new_column')

    def test_add_column_with_default_value(self):
        t = self.registry.migration.table('test')
        t.column().add(Column('new_column', Integer, default=100))
        t.column('new_column')

    def test_add_column_with_default_str_value(self):
        t = self.registry.migration.table('test')
        t.column().add(Column('new_column', Integer, default='100'))
        t.column('new_column')

    def test_add_column_with_default_callable_value(self):

        def get_val():
            return 100

        self.fill_test_table()
        t = self.registry.migration.table('test')
        t.column().add(Column('new_column', Integer, default=get_val))
        t.column('new_column')
        res = [x for x in self.registry.execute(
            "select count(*) from test where new_column is null")][0][0]
        self.assertEqual(res, 0)

    def test_add_column_in_filled_table_with_default_value(self):
        self.fill_test_table()
        t = self.registry.migration.table('test')
        t.column().add(Column('new_column', Integer, default=100))
        t.column('new_column')
        res = [x for x in self.registry.execute(
            "select count(*) from test where new_column is null")][0][0]
        self.assertEqual(res, 0)

    def test_add_not_null_column_in_filled_table_with_default_value(self):
        self.fill_test_table()
        t = self.registry.migration.table('test')
        t.column().add(Column('new_column', Integer, nullable=False,
                       server_default="1"))
        t.column('new_column')

    def test_add_unique_constraint_on_good_table(self):
        self.fill_test_table()
        t = self.registry.migration.table('test')
        t.unique('unique_constraint').add(t.column('other'))
        self.registry.Test.insert(other='One entry')
        with self.assertRaises(IntegrityError):
            self.registry.Test.insert(other='One entry')

    def test_add_unique_constraint_on_not_unique_column(self):
        Test = self.registry.Test
        vals = [{'other': 'test'} for x in range(10)]
        Test.multi_insert(*vals)
        t = self.registry.migration.table('test')
        t.unique(None).add(t.column('other'))
        self.registry.Test.insert(other='One entry')
        self.registry.Test.insert(other='One entry')

    def test_drop_table(self):
        self.registry.migration.table('test').drop()
        with self.assertRaises(MigrationException):
            self.registry.migration.table('Test')

    def test_drop_column(self):
        t = self.registry.migration.table('test')
        t.column('other').drop()
        with self.assertRaises(MigrationException):
            t.column('other')

    def test_alter_table_name(self):
        t = self.registry.migration.table('test')
        t.alter(name='othername')
        self.registry.migration.table('othername')
        with self.assertRaises(MigrationException):
            self.registry.migration.table('test')

    def test_alter_column_name(self):
        t = self.registry.migration.table('test')
        t.column('other').alter(name='othername')
        t.column('othername')
        with self.assertRaises(MigrationException):
            t.column('other')

    def test_alter_column_nullable(self):
        t = self.registry.migration.table('test')
        c = t.column('other').alter(nullable=False)
        self.assertFalse(c.nullable())

    def test_alter_column_nullable_in_filled_table(self):
        t = self.registry.migration.table('test')
        t.column().add(Column('new_column', Integer))
        self.fill_test_table()
        c = t.column('new_column').alter(nullable=False)
        # the column doesn't change of nullable to not lock the migration
        self.assertTrue(c.nullable())

    def test_alter_column_default(self):
        t = self.registry.migration.table('test')
        c = t.column('other').alter(server_default='test')
        self.assertEqual(c.server_default(), "'test'::character varying")

    def test_index(self):
        t = self.registry.migration.table('test')
        t.index().add(t.column('integer'))
        t.index(t.column('integer')).drop()

    def test_alter_column_type(self):
        t = self.registry.migration.table('test')
        c = t.column('other').alter(type_=TEXT)
        self.assertEqual(c.type().__class__, TEXT)

    def test_alter_column_primary_key(self):
        t = self.registry.migration.table('test')
        t.primarykey().drop().add(t.column('other'))

    def test_alter_column_foreign_key(self):
        t = self.registry.migration.table('test')
        t.foreign_key('my_fk').add(
            ['other'], [self.registry.System.Blok.name])
        t.foreign_key('my_fk').drop()

    def test_constraint_unique(self):
        t = self.registry.migration.table('test')
        t.unique(name='test_unique_constraint').add(t.column('other'))
        t.unique(name='test_unique_constraint').drop()

    def test_constraint_check(self):
        t = self.registry.migration.table('test')
        Test = self.registry.Test
        t.check('test').add(Test.other != 'test')
        # particuliar case of check constraint
        t.check('anyblok_ck_test_test').drop()

    def test_detect_under_noautocommit_flag(self):
        with self.cnx() as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) NOT NULL
                );""")
        self.registry.migration.detect_changed()
        self.registry.migration.withoutautomigration = True
        with self.assertRaises(MigrationException):
            self.registry.migration.detect_changed()

    def test_detect_column_added(self):
        # Remove a column on the table force the detection to found new column
        # which is existing in metadata but not in table
        with self.cnx() as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(integer INT PRIMARY KEY NOT NULL);""")
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Add test.other"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertFalse(report.log_has("Add test.other"))

    def test_detect_table_removed(self):
        with self.cnx() as conn:
            conn.execute(
                """CREATE TABLE test2(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64),
                    other2 CHAR(64)
                );""")

        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Drop Table test2"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Drop Table test2"))

    def test_detect_table_removed_with_reinit_column(self):
        with self.cnx() as conn:
            conn.execute(
                """CREATE TABLE test2(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64),
                    other2 CHAR(64)
                );""")

        self.registry.migration.reinit_tables = True
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Drop Table test2"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertFalse(report.log_has("Drop Table test2"))

    def test_detect_table_removed_with_reinit_all(self):
        with self.cnx() as conn:
            conn.execute(
                """CREATE TABLE test2(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64),
                    other2 CHAR(64)
                );""")

        self.registry.migration.reinit_all = True
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Drop Table test2"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertFalse(report.log_has("Drop Table test2"))

    def test_detect_column_removed(self):
        with self.cnx() as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64),
                    other2 CHAR(64)
                );""")
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Drop Column test.other2"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Drop Column test.other2"))

    def test_detect_column_removed_with_reinit_column(self):
        with self.cnx() as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64),
                    other2 CHAR(64)
                );""")
        self.registry.migration.reinit_columns = True
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Drop Column test.other2"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertFalse(report.log_has("Drop Column test.other2"))

    def test_detect_column_removed_with_reinit_all(self):
        with self.cnx() as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64),
                    other2 CHAR(64)
                );""")
        self.registry.migration.reinit_all = True
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Drop Column test.other2"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertFalse(report.log_has("Drop Column test.other2"))

    def test_detect_not_nullable_column_removed(self):
        with self.cnx() as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64),
                    other2 CHAR(64) NOT NULL
                );""")
        report = self.registry.migration.detect_changed()
        self.assertFalse(report.log_has("Drop Column test.other2"))
        self.assertTrue(report.log_has("Drop Column test.other2 (not null)"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Drop Column test.other2"))
        self.assertFalse(report.log_has("Drop Column test.other2 (not null)"))

    def test_detect_nullable(self):
        with self.cnx() as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) NOT NULL
                );""")
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Alter test.other"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertFalse(report.log_has("Alter test.other"))

    @skipIf(skipAlembicVersion, "Alembic doesn't implement yet")
    def test_detect_m2m_primary_key(self):
        with self.cnx() as conn:
            conn.execute("DROP TABLE reltable")
            conn.execute(
                """CREATE TABLE reltable (
                    idmodel1 INT,
                    idmodel2 INT,
                    FOREIGN KEY (idmodel1) REFERENCES testm2m1 (idmodel1),
                    FOREIGN KEY (idmodel2) REFERENCES testm2m2 (idmodel2)
                );""")
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Alter reltable.idmodel1"))
        self.assertTrue(report.log_has("Alter reltable.idmodel2"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertFalse(report.log_has("Alter reltable.idmodel1"))
        self.assertFalse(report.log_has("Alter reltable.idmodel2"))

    def test_detect_server_default(self):
        with self.cnx() as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) DEFAULT 9.99
                );""")
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Alter test.other"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertFalse(report.log_has("Alter test.other"))

    def test_detect_drop_index(self):
        with self.cnx() as conn:
            conn.execute("""CREATE INDEX other_idx ON test (other);""")
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Drop index other_idx on test"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Drop index other_idx on test"))

    def test_detect_drop_anyblok_index(self):
        with self.cnx() as conn:
            conn.execute(
                """CREATE INDEX anyblok_ix_test__other ON test (other);""")
        report = self.registry.migration.detect_changed()
        self.assertTrue(
            report.log_has("Drop index anyblok_ix_test__other on test"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertFalse(
            report.log_has("Drop index anyblok_ix_test__other on test"))

    def test_detect_drop_index_with_reinit_indexes(self):
        with self.cnx() as conn:
            conn.execute("""CREATE INDEX other_idx ON test (other);""")
        self.registry.migration.reinit_indexes = True
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Drop index other_idx on test"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertFalse(report.log_has("Drop index other_idx on test"))

    def test_detect_drop_index_with_reinit_all(self):
        with self.cnx() as conn:
            conn.execute("""CREATE INDEX other_idx ON test (other);""")
        self.registry.migration.reinit_all = True
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Drop index other_idx on test"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertFalse(report.log_has("Drop index other_idx on test"))

    def test_detect_type(self):
        with self.cnx() as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other INT
                );""")
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Alter test.other"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertFalse(report.log_has("Alter test.other"))

    @skipIf(skipAlembicVersion, "Alembic doesn't implement yet")
    def test_detect_primary_key(self):
        with self.cnx() as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT NOT NULL,
                    other CHAR(64) PRIMARY KEY NOT NULL
                );""")
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Alter test.integer"))
        self.assertTrue(report.log_has("Alter test.other"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertFalse(report.log_has("Alter test.integer"))
        self.assertFalse(report.log_has("Alter test.other"))

    def test_detect_add_foreign_key(self):
        with self.cnx() as conn:
            conn.execute("DROP TABLE testfk")
            conn.execute(
                """CREATE TABLE testfk(
                    integer INT PRIMARY KEY NOT NULL,
                    other INT
                );""")
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has(
            "Add Foreign keys on (testfk.other) => (testfktarget.integer)"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertFalse(report.log_has(
            "Add Foreign keys on (testfk.other) => (testfktarget.integer)"))

    def test_detect_drop_foreign_key(self):
        with self.cnx() as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) references system_blok(name)
                );""")
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has(
            "Drop Foreign keys on test.other => system_blok.name"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has(
            "Drop Foreign keys on test.other => system_blok.name"))

    def test_detect_drop_anyblok_foreign_key(self):
        with self.cnx() as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64)
                        CONSTRAINT anyblok_fk_test__other_on_system_blok__name
                        references system_blok(name)
                );""")
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has(
            "Drop Foreign keys on test.other => system_blok.name"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertFalse(report.log_has(
            "Drop Foreign keys on test.other => system_blok.name"))

    def test_detect_drop_foreign_key_with_reinit_constraint(self):
        with self.cnx() as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) references system_blok(name)
                );""")
        self.registry.migration.reinit_constraints = True
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has(
            "Drop Foreign keys on test.other => system_blok.name"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertFalse(report.log_has(
            "Drop Foreign keys on test.other => system_blok.name"))

    def test_detect_drop_foreign_key_with_reinit_all(self):
        with self.cnx() as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) references system_blok(name)
                );""")
        self.registry.migration.reinit_all = True
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has(
            "Drop Foreign keys on test.other => system_blok.name"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertFalse(report.log_has(
            "Drop Foreign keys on test.other => system_blok.name"))

    def test_detect_drop_column_with_foreign_key(self):
        with self.cnx() as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64),
                    other2 CHAR(64)
                    CONSTRAINT anyblok_fk_test__other2_on_system_blok__name
                    references system_blok(name)
                );""")
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has(
            "Drop Foreign keys on test.other2 => system_blok.name"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertFalse(report.log_has(
            "Drop Foreign keys on test.other2 => system_blok.name"))

    def test_detect_add_unique_constrainte(self):
        with self.cnx() as conn:
            conn.execute("DROP TABLE testunique")
            conn.execute(
                """CREATE TABLE testunique(
                    integer INT  PRIMARY KEY NOT NULL,
                    other CHAR(64)
                );""")
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has(
            "Add unique constraint on testunique (other)"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertFalse(report.log_has(
            "Add unique constraint on testunique (other)"))

    def test_detect_drop_unique_constraint(self):
        with self.cnx() as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) CONSTRAINT unique_other UNIQUE
                );""")
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Drop constraint unique_other on test"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Drop constraint unique_other on test"))

    def test_detect_drop_unique_anyblok_constraint(self):
        with self.cnx() as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) CONSTRAINT anyblok_uq_test__other UNIQUE
                );""")
        report = self.registry.migration.detect_changed()
        self.assertTrue(
            report.log_has("Drop constraint anyblok_uq_test__other on test"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertFalse(
            report.log_has("Drop constraint anyblok_uq_test__other on test"))

    def test_detect_drop_unique_constraint_with_reinit_constraints(self):
        with self.cnx() as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) CONSTRAINT unique_other UNIQUE
                );""")

        self.registry.migration.reinit_constraints = True
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Drop constraint unique_other on test"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertFalse(
            report.log_has("Drop constraint unique_other on test"))

    def test_detect_drop_unique_constraint_with_reinit_all(self):
        with self.cnx() as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) CONSTRAINT unique_other UNIQUE
                );""")

        self.registry.migration.reinit_all = True
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Drop constraint unique_other on test"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertFalse(
            report.log_has("Drop constraint unique_other on test"))

    @skipIf(skipAlembicVersion, "Alembic doesn't implement yet")
    def test_detect_drop_check_constraint(self):
        with self.cnx() as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) CONSTRAINT ck_other CHECK (other != 'test')
                );""")
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Drop constraint ck_other on test"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Drop constraint ck_other on test"))

    @skipIf(skipAlembicVersion, "Alembic doesn't implement yet")
    def test_detect_drop_check_anyblok_constraint(self):
        with self.cnx() as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) CONSTRAINT anyblok_ck_test_check
                        CHECK (other != 'test')
                );""")
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has(
            "Drop constraint anyblok_ck_test_check on test"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertFalse(report.log_has(
            "Drop constraint anyblok_ck_test_check on test"))

    @skipIf(skipAlembicVersion, "Alembic doesn't implement yet")
    def test_detect_drop_check_constraint_with_reinit_constraint(self):
        with self.cnx() as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) CONSTRAINT ck_other CHECK (other != 'test')
                );""")
        self.registry.migration.reinit_constraints = True
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Drop constraint ck_other on test"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertFalse(report.log_has("Drop constraint ck_other on test"))

    @skipIf(skipAlembicVersion, "Alembic doesn't implement yet")
    def test_detect_drop_check_constraint_with_reinit_all(self):
        with self.cnx() as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) CONSTRAINT ck_other CHECK (other != 'test')
                );""")
        self.registry.migration.reinit_all = True
        report = self.registry.migration.detect_changed()
        self.assertTrue(report.log_has("Drop constraint ck_other on test"))
        report.apply_change()
        report = self.registry.migration.detect_changed()
        self.assertFalse(report.log_has("Drop constraint ck_other on test"))

    def test_savepoint(self):
        Test = self.registry.Test
        self.fill_test_table()
        self.registry.migration.savepoint('test')
        self.fill_test_table()
        self.assertEqual(Test.query().count(), 20)
        self.registry.migration.rollback_savepoint('test')
        self.assertEqual(Test.query().count(), 10)
        self.registry.migration.release_savepoint('test')

    def test_savepoint_without_rollback(self):
        self.registry.migration.savepoint('test')
        self.registry.migration.release_savepoint('test')
        with self.assertRaises(InternalError):
            self.registry.migration.rollback_savepoint('test')



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Pierre Verkest <pverkest@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from sqlalchemy import Table, Column, ForeignKeyConstraint
from sqlalchemy.orm import (relationships, backref, relationship, base,
                            attributes)
from sqlalchemy.schema import Column as SA_Column
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy import exc as sa_exc, util
from .field import Field, FieldException
from .mapper import ModelAdapter, ModelAttribute, ModelRepr
from anyblok.common import anyblok_column_prefix


class RelationshipProperty(relationships.RelationshipProperty):

    def __init__(self, *args, **kwargs):
        self.relationship_field = kwargs.pop('relationship_field')
        super(RelationshipProperty, self).__init__(*args, **kwargs)

    def _generate_backref(self):  # noqa
        """Interpret the 'backref' instruction to create a
        :func:`.relationship` complementary to this one."""

        if self.parent.non_primary:
            return
        if self.backref is not None and not self.back_populates:
            if isinstance(self.backref, util.string_types):
                backref_key, kwargs = self.backref, {}
            else:
                backref_key, kwargs = self.backref
            mapper = self.mapper.primary_mapper()

            check = set(mapper.iterate_to_root()).\
                union(mapper.self_and_descendants)
            for m in check:
                if m.has_property(backref_key):
                    raise sa_exc.ArgumentError(
                        "Error creating backref "
                        "'%s' on relationship '%s': property of that "
                        "name exists on mapper '%s'" %
                        (backref_key, self, m))

            # determine primaryjoin/secondaryjoin for the
            # backref.  Use the one we had, so that
            # a custom join doesn't have to be specified in
            # both directions.
            if self.secondary is not None:
                # for many to many, just switch primaryjoin/
                # secondaryjoin.   use the annotated
                # pj/sj on the _join_condition.
                pj = kwargs.pop(
                    'primaryjoin',
                    self._join_condition.secondaryjoin_minus_local)
                sj = kwargs.pop(
                    'secondaryjoin',
                    self._join_condition.primaryjoin_minus_local)
            else:
                pj = kwargs.pop(
                    'primaryjoin',
                    self._join_condition.primaryjoin_reverse_remote)
                sj = kwargs.pop('secondaryjoin', None)
                if sj:
                    raise sa_exc.InvalidRequestError(
                        "Can't assign 'secondaryjoin' on a backref "
                        "against a non-secondary relationship."
                    )

            foreign_keys = kwargs.pop('foreign_keys',
                                      self._user_defined_foreign_keys)
            parent = self.parent.primary_mapper()
            kwargs.setdefault('viewonly', self.viewonly)
            kwargs.setdefault('post_update', self.post_update)
            kwargs.setdefault('passive_updates', self.passive_updates)
            self.back_populates = backref_key
            _relationship = RelationshipProperty2(
                parent, self.secondary,
                pj, sj,
                foreign_keys=foreign_keys,
                back_populates=self.key,
                relationship_field=self.relationship_field,
                **kwargs)

            mapper._configure_property(backref_key, _relationship)

        if self.back_populates:
            self._add_reverse_property(self.back_populates)


def register_descriptor(class_, key, comparator=None,
                        parententity=None, doc=None, relationship_field=None):
    manager = base.manager_of_class(class_)
    descriptor = relationship_field.InstrumentedAttribute(
        class_, key, comparator=comparator, parententity=parententity,
        relationship_field=relationship_field)
    descriptor.__doc__ = doc
    manager.instrument_attribute(key, descriptor)
    return descriptor


class RelationshipProperty2(relationships.RelationshipProperty):

    def __init__(self, *args, **kwargs):
        self.relationship_field = kwargs.pop('relationship_field')
        super(RelationshipProperty2, self).__init__(*args, **kwargs)

    def instrument_class(self, mapper):
        register_descriptor(
            mapper.class_,
            self.key,
            comparator=self.comparator_factory(self, mapper),
            parententity=mapper,
            doc=self.doc,
            relationship_field=self.relationship_field,
        )


class RelationShipList:  # don't inherit list

    def append(self, x):
        res = super(RelationShipList, self).append(x)
        self.relationship_field_append_value(x)
        return res

    def extend(self, L):
        res = super(RelationShipList, self).extend(L)
        for el in L:
            self.relationship_field_append_value(el)

        return res

    def insert(self, i, x):
        res = super(RelationShipList, self).insert(i, x)
        self.relationship_field_append_value(x)
        return res

    def remove(self, x):
        self.relationship_field_remove_value(x)
        return super(RelationShipList, self).remove(x)

    def pop(self, *args, **kwargs):
        res = super(RelationShipList, self).pop(*args, **kwargs)
        self.relationship_field_remove_value(res)
        return res

    def clear(self):
        for x in self:
            self.relationship_field_remove_value(x)

        return super(RelationShipList, self).clear()


class RelationShip(Field):
    """ RelationShip class

    The RelationShip class is used to define the type of SQL field Declarations

    Add a new relation ship type::

        @Declarations.register(Declarations.RelationShip)
        class Many2one:
            pass

    the relationship column are forbidden because the model can be used on
    the model
    """

    def __init__(self, *args, **kwargs):
        self.forbid_instance(RelationShip)
        if 'model' in kwargs:
            self.model = ModelAdapter(kwargs.pop('model'))
        else:
            raise FieldException("model is required attribut")

        super(RelationShip, self).__init__(*args, **kwargs)

        if 'info' not in self.kwargs:
            self.kwargs['info'] = {}

        self.kwargs['info']['remote_model'] = self.model.model_name
        self.backref_properties = {}

    def autodoc_get_properties(self):
        res = super(RelationShip, self).autodoc_get_properties()
        res['model'] = self.model
        return res

    def apply_instrumentedlist(self, registry, namespace, fieldname):
        """ Add the InstrumentedList class to replace List class as result
        of the query

        :param registry: current registry
        """
        self.kwargs['collection_class'] = registry.InstrumentedList
        self.backref_properties['collection_class'] = registry.InstrumentedList

    def define_backref_properties(self, registry, namespace, properties):
        """ Add in the backref_properties, new property for the backref

        :param registry: current registry
        :param namespace: name of the model
        :param properties: properties known of the model
        """
        pass

    def format_backref(self, registry, namespace, fieldname, properties):
        """ Create the real backref, with the backref string and the
        backref properties

        :param registry: current registry
        :param namespace: name of the model
        :param fieldname: name of the field
        :param properties: properties known of the model
        """
        _backref = self.kwargs.get('backref')
        if not _backref:
            return

        if isinstance(_backref, (list, tuple)):
            _backref, backref_properties = _backref
            self.backref_properties.update(backref_properties)

        self.define_backref_properties(registry, namespace, properties)

        if self.backref_properties:
            self.kwargs['backref'] = backref(_backref,
                                             **self.backref_properties)
            mapper = ModelAttribute(self.model.model_name, _backref)
            if not mapper.is_declared(registry):
                mapper.add_fake_relationship(registry, namespace, fieldname)

    def get_relationship_cls(self):
        return relationship

    def get_sqlalchemy_mapping(self, registry, namespace, fieldname,
                               properties):
        """ Return the instance of the real field

        :param registry: current registry
        :param namespace: name of the model
        :param fieldname: name of the field
        :param properties: properties known of the model
        :rtype: sqlalchemy relation ship instance
        """
        self.model.check_model(registry)
        self.format_label(fieldname)
        self.kwargs['info']['label'] = self.label
        self.kwargs['info']['rtype'] = self.__class__.__name__
        self.apply_instrumentedlist(registry, namespace, fieldname)
        self.format_backref(registry, namespace, fieldname, properties)
        return self.get_relationship_cls()(
            self.model.tablename(registry), **self.kwargs)

    def must_be_declared_as_attr(self):
        """ Return True, because it is a relationship """
        return True

    def init_expire_attributes(self, registry, namespace, fieldname):
        """Init dict of expiration properties

        :param registry: current registry
        :param namespace: name of the model
        :param fieldname: name of the field
        """
        if namespace not in registry.expire_attributes:
            registry.expire_attributes[namespace] = {}

        if fieldname not in registry.expire_attributes[namespace]:
            registry.expire_attributes[namespace][fieldname] = set()


class RelationShipListMany2One:

    def relationship_field_append_value(self, value):
        for model_field, rfield in self.relationship_fied.link_between_columns:
            self.relationship_fied.apply_value_to(
                value, model_field, getattr(value, self.fieldname), rfield)

    def relationship_field_remove_value(self, value):
        for model_field, rfield in self.relationship_fied.link_between_columns:
            setattr(value, anyblok_column_prefix + model_field, None)


class Many2One(RelationShip):
    """ Define a relationship attribute on the model

    ::

        @register(Model)
        class TheModel:

            relationship = Many2One(label="The relationship",
                                    model=Model.RemoteModel,
                                    remote_columns="The remote column",
                                    column_names="The column which have the "
                                                 "foreigh key",
                                    nullable=True,
                                    unique=False,
                                    one2many="themodels")

    If the ``remote_columns`` are not define then, the system takes the primary
    key of the remote model

    If the column doesn't exist, the column will be created. Use the
    nullable option.
    If the name is not filled, the name is "'remote table'_'remote colum'"

    :param model: the remote model
    :param remote_columns: the column name on the remote model
    :param column_names: the column on the model which have the foreign key
    :param nullable: If the column_names is nullable
    :param unique: If True, add the unique constraint on the column
    :param one2many: create the one2many link with this many2one
    """
    use_hybrid_property = True

    def __init__(self, **kwargs):
        super(Many2One, self).__init__(**kwargs)

        self._remote_columns = None
        if 'remote_columns' in kwargs:
            self._remote_columns = self.kwargs.pop('remote_columns')
            if not isinstance(self._remote_columns, (list, tuple)):
                self._remote_columns = [self._remote_columns]

        self.nullable = True
        if 'nullable' in kwargs:
            self.nullable = self.kwargs.pop('nullable')
            self.kwargs['info']['nullable'] = self.nullable

        self.unique = False
        if 'unique' in kwargs:
            self.unique = self.kwargs.pop('unique')
            self.kwargs['info']['unique'] = self.unique

        if 'one2many' in kwargs:
            self.kwargs['backref'] = self.kwargs.pop('one2many')
            self.kwargs['info']['remote_name'] = self.kwargs['backref']

        self._column_names = None
        if 'column_names' in kwargs:
            self._column_names = self.kwargs.pop('column_names')
            if not isinstance(self._column_names, (list, tuple)):
                self._column_names = [self._column_names]

        self.foreign_key_options = self.kwargs.pop('foreign_key_options', {})
        self.cascade = self.kwargs.pop('cascade', 'save-update, merge')

    def autodoc_get_properties(self):
        res = super(Many2One, self).autodoc_get_properties()
        res['_remote_columns'] = self._remote_columns
        res['_column_names'] = self._column_names
        res['unique'] = self.unique
        return res

    def update_local_and_remote_columns_names(self, registry, namespace):
        if self._remote_columns is None:
            self.remote_columns = self.model.primary_keys(registry)
        else:
            self.remote_columns = [ModelAttribute(self.model.model_name, x)
                                   for x in self._remote_columns]

        self.kwargs['info']['remote_columns'] = [str(x)
                                                 for x in self.remote_columns]

        if self._column_names is None:
            model = ModelRepr(namespace)
            self.column_names = model.foreign_keys_for(
                registry, self.model.model_name)
            if not self.column_names:
                self.column_names = []
                for x in self.remote_columns:
                    cname = x.get_fk_name(registry).replace('.', '_')
                    self.column_names.append(ModelAttribute(namespace, cname))
        else:
            self.column_names = [ModelAttribute(namespace, x)
                                 for x in self._column_names]

    def get_property(self, registry, namespace, fieldname, properties):
        """Return the property of the field

        :param registry: current registry
        :param namespace: name of the model
        :param fieldname: name of the field
        :param properties: properties known to the model
        """
        res = super(Many2One, self).get_property(
            registry, namespace, fieldname, properties)
        # force the info value in hybrid_property because since SQLAlchemy
        # 1.1.* the info is not propagate
        res.info = self.kwargs['info']
        return res

    def add_expire_attributes(self, registry, namespace, fieldname, cname):
        self.init_expire_attributes(registry, namespace, cname)
        registry.expire_attributes[namespace][cname].add((fieldname,))
        if self.kwargs.get('backref'):
            backref = self.kwargs['backref']
            if isinstance(backref, (list, tuple)):
                backref = backref[0]

            registry.expire_attributes[namespace][cname].add(
                (fieldname, backref))

    def update_properties(self, registry, namespace, fieldname, properties):
        """ Create the column which has the foreign key if the column doesn't
        exist

        :param registry: the registry which load the relationship
        :param namespace: the name space of the model
        :param fieldname: fieldname of the relationship
        :param propertie: the properties known
        """
        add_fksc = False
        self.link_between_columns = []
        self.model.check_model(registry)
        self.update_local_and_remote_columns_names(registry, namespace)
        if fieldname in [x.attribute_name for x in self.column_names]:
            raise FieldException("The column_names and the fieldname %r are "
                                 "the same, please choose another "
                                 "column_names" % fieldname)

        self.kwargs['info']['local_columns'] = ', '.join(
            str(x) for x in self.column_names)
        remote_types = {x.attribute_name: x.native_type(registry)
                        for x in self.remote_columns}
        remote_columns = {x.attribute_name: x
                          for x in self.remote_columns}
        for cname in self.column_names:
            if cname.is_declared(registry):
                del remote_types[cname.get_fk_column(registry)]

        col_names = []
        fk_names = []
        for cname in self.column_names:
            self.add_expire_attributes(registry, namespace, fieldname,
                                       cname.attribute_name)
            if not cname.is_declared(registry):
                rc, remote_type = self.get_column_information(
                    registry, cname, remote_types)
                cname.add_fake_column(registry)
                foreign_key = remote_columns[rc].get_fk_name(registry)
                self.create_column(cname, remote_type, foreign_key, properties)
                add_fksc = True
                fk_name = remote_columns[rc]
            else:
                fk_name = cname.get_fk_mapper(registry)

            col_names.append(cname.attribute_name)
            fk_names.append(fk_name.get_fk_name(registry))
            self.link_between_columns.append((cname.attribute_name,
                                              fk_name.attribute_name))

        if namespace == self.model.model_name:
            self.kwargs['remote_side'] = [
                properties[anyblok_column_prefix + x.attribute_name]
                for x in self.remote_columns]

        if (len(self.column_names) > 1 or add_fksc) and col_names and fk_names:
            properties['add_in_table_args'].append(
                ForeignKeyConstraint(col_names, fk_names,
                                     **self.foreign_key_options))

    def get_column_information(self, registry, cname, remote_types):
        if len(remote_types) == 1:
            rc = [x for x in remote_types][0]
            return rc, remote_types[rc]
        else:
            rc = cname.get_fk_column(registry)
            if rc is None:
                rc = cname.attribute_name[len(
                    self.model.tablename(registry)) + 1:]

            if rc in remote_types:
                return rc, remote_types[rc]
            else:
                cname.get_fk_column(registry)
                raise FieldException("Can not create the local "
                                     "column %r" % cname.attribute_name)

    def apply_instrumentedlist(self, registry, namespace, fieldname):
        """ Add the InstrumentedList class to replace List class as result
        of the query

        :param registry: current registry
        """
        properties = {
            'fieldname': fieldname, 'relationship_fied': self}
        InstrumentedList = type(
            'InstrumentedList', (RelationShipListMany2One, RelationShipList,
                                 registry.InstrumentedList), properties)
        self.backref_properties['collection_class'] = InstrumentedList
        cascade = self.cascade
        if self.foreign_key_options.get('ondelete') == 'cascade':
            cascade += ', delete'
        self.backref_properties['cascade'] = cascade

    def create_column(self, cname, remote_type, foreign_key, properties):

        def wrapper(cls):
            return SA_Column(
                cname.attribute_name,
                remote_type,
                nullable=self.nullable,
                unique=self.unique,
                info=dict(label=self.label, foreign_key=foreign_key))

        properties[(anyblok_column_prefix +
                    cname.attribute_name)] = declared_attr(wrapper)
        properties['loaded_columns'].append(cname.attribute_name)
        properties['hybrid_property_columns'].append(cname.attribute_name)
        properties[cname.attribute_name] = hybrid_property(
            self.wrap_getter_column(cname.attribute_name),
            super(Many2One, self).wrap_setter_column(cname.attribute_name),
            expr=self.wrap_expr_column(cname.attribute_name))

    def get_sqlalchemy_mapping(self, registry, namespace, fieldname,
                               properties):
        """ Create the relationship

        :param registry: the registry which load the relationship
        :param namespace: the name space of the model
        :param fieldname: fieldname of the relationship
        :param propertie: the properties known
        :rtype: Many2One relationship
        """
        self.kwargs['foreign_keys'] = '[%s]' % ', '.join(
            [x.get_fk_name(registry) for x in self.column_names])
        return super(Many2One, self).get_sqlalchemy_mapping(
            registry, namespace, fieldname, properties)

    def apply_value_to(self, model_self, model_field, remote_self,
                       remote_field):
        if remote_self:
            value = getattr(remote_self,
                            anyblok_column_prefix + remote_field)
        else:
            value = None

        setattr(model_self, anyblok_column_prefix + model_field, value)

    def wrap_setter_column(self, fieldname):
        attr_name = anyblok_column_prefix + fieldname

        def setter_column(model_self, value):
            res = setattr(model_self, attr_name, value)
            for model_field, rfield in self.link_between_columns:
                self.apply_value_to(model_self, model_field, value, rfield)

            return res

        return setter_column


class InstrumentedAttribute_O2O(attributes.InstrumentedAttribute):

    def __init__(self, *args, **kwargs):
        self.relationship_field = kwargs.pop('relationship_field')
        super(InstrumentedAttribute_O2O, self).__init__(*args, **kwargs)

    def __set__(self, instance, value):
        call_super = False
        if value:
            for cname, fname in self.relationship_field.link_between_columns:
                if instance:
                    if getattr(instance, fname):
                        setattr(value, cname, getattr(instance, fname))
                    else:
                        call_super = True
                else:
                    setattr(value, cname, instance)

        else:
            call_super = True

        if call_super:
            super(InstrumentedAttribute_O2O, self).__set__(instance, value)


class One2One(Many2One):
    """ Define a relationship attribute on the model

    ::

        @register(Model)
        class TheModel:

            relationship = One2One(label="The relationship",
                                   model=Model.RemoteModel,
                                   remote_columns="The remote column",
                                   column_names="The column which have the "
                                               "foreigh key",
                                   nullable=False,
                                   backref="themodels")

    If the remote_columns are not define then, the system take the primary key
    of the remote model

    If the column doesn't exist, then the column will be create. Use the
    nullable option.
    If the name is not filled then the name is "'remote table'_'remote colum'"

    :param model: the remote model
    :param remote_columns: the column name on the remote model
    :param column_names: the column on the model which have the foreign key
    :param nullable: If the column_names is nullable
    :param backref: create the one2one link with this one2one
    """

    InstrumentedAttribute = InstrumentedAttribute_O2O

    def __init__(self, **kwargs):
        super(One2One, self).__init__(**kwargs)

        if 'backref' not in kwargs:
            raise FieldException("backref is a required argument")

        if 'one2many' in kwargs:
            raise FieldException("Unknow argmument 'one2many'")

        self.kwargs['info']['remote_name'] = self.kwargs['backref']

    def define_backref_properties(self, registry, namespace, properties):
        """ Add option uselist = False

        :param registry: the registry which load the relationship
        :param namespace: the name space of the model
        :param propertie: the properties known
        """
        self.backref_properties.update({'uselist': False})

    def apply_instrumentedlist(self, registry, namespace, fieldname):
        """ Add the InstrumentedList class to replace List class as result
        of the query

        :param registry: current registry
        """

    def get_relationship_cls(self):
        self.kwargs['relationship_field'] = self
        return RelationshipProperty


class Many2Many(RelationShip):
    """ Define a relationship attribute on the model

    ::

        @register(Model)
        class TheModel:

            relationship = Many2Many(label="The relationship",
                                     model=Model.RemoteModel,
                                     join_table="many2many table",
                                     remote_columns="The remote column",
                                     m2m_remote_columns="Name in many2many"
                                     local_columns="local primary key"
                                     m2m_local_columns="Name in many2many"
                                     many2many="themodels")

    if the join_table is not defined, then the table join is
        "join\_'local table'_and\_'remote table'"

    .. warning::

        The join_table must be filled when the declaration of the
        Many2Many is done in a Mixin

    If the remote_columns are not define then, the system take the primary key
    of the remote model

    if the local_columns are not define the take the primary key of the local
        model

    :param model: the remote model
    :param join_table: the many2many table to join local and remote models
    :param remote_columns: the column name on the remote model
    :param m2m_remote_columns: the column name to remote model in m2m table
    :param local_columns: the column on the model
    :param m2m_local_columns: the column name to local model in m2m table
    :param many2many: create the opposite many2many on the remote model
    """

    def __init__(self, **kwargs):
        super(Many2Many, self).__init__(**kwargs)

        self.join_table = self.kwargs.pop('join_table', None)
        self.remote_columns = self.kwargs.pop('remote_columns', None)
        if self.remote_columns and not isinstance(self.remote_columns,
                                                  (list, tuple)):
            self.remote_columns = [self.remote_columns]

        self.m2m_remote_columns = self.kwargs.pop('m2m_remote_columns', None)
        if self.m2m_remote_columns and not isinstance(self.m2m_remote_columns,
                                                      (list, tuple)):
            self.m2m_remote_columns = [self.m2m_remote_columns]

        self.local_columns = self.kwargs.pop('local_columns', None)
        if self.local_columns and not isinstance(self.local_columns,
                                                 (list, tuple)):
            self.local_columns = [self.local_columns]

        self.m2m_local_columns = self.kwargs.pop('m2m_local_columns', None)
        if self.m2m_local_columns and not isinstance(self.m2m_local_columns,
                                                     (list, tuple)):
            self.m2m_local_columns = [self.m2m_local_columns]

        self.kwargs['backref'] = self.kwargs.pop('many2many', None)
        self.kwargs['info']['remote_name'] = self.kwargs['backref']

    def autodoc_get_properties(self):
        res = super(Many2Many, self).autodoc_get_properties()
        res['join table'] = self.join_table
        res['remote_columns'] = self.remote_columns
        res['m2m_remote_columns'] = self.m2m_remote_columns
        res['local_columns'] = self.local_columns
        res['m2m_local_columns'] = self.m2m_local_columns
        return res

    def get_m2m_columns(self, registry, columns, m2m_columns):
        if m2m_columns is None:
            m2m_columns = [x.get_fk_name(registry).replace('.', '_')
                           for x in columns]

        cols = []
        col_names = []
        ref_cols = []
        for i, column in enumerate(m2m_columns):
            sqltype = columns[i].native_type(registry)
            foreignkey = columns[i].get_fk_name(registry)
            cols.append(Column(column, sqltype, primary_key=True))
            col_names.append(column)
            ref_cols.append(foreignkey)

        return cols, ForeignKeyConstraint(col_names, ref_cols)

    def get_local_and_remote_columns(self, registry):
        if not self.local_columns:
            local_columns = self.local_model.primary_keys(registry)
        else:
            local_columns = [ModelAttribute(self.local_model.model_name, x)
                             for x in self.local_columns]

        if not self.remote_columns:
            remote_columns = self.model.primary_keys(registry)
        else:
            remote_columns = [ModelAttribute(self.model.model_name, x)
                              for x in self.remote_columns]

        self.kwargs['info']['local_columns'] = [x.attribute_name
                                                for x in local_columns]
        self.kwargs['info']['remote_columns'] = [x.attribute_name
                                                 for x in remote_columns]

        return local_columns, remote_columns

    def get_sqlalchemy_mapping(self, registry, namespace, fieldname,
                               properties):
        """ Create the relationship

        :param registry: the registry which load the relationship
        :param namespace: the name space of the model
        :param fieldname: fieldname of the relationship
        :param properties: the properties known
        :rtype: Many2One relationship
        """
        self.model.check_model(registry)
        self.local_model = ModelRepr(namespace)
        local_columns, remote_columns = self.get_local_and_remote_columns(
            registry)
        join_table = self.join_table
        if self.join_table is None:
            join_table = 'join_%s_and_%s' % (
                self.local_model.tablename(registry),
                self.model.tablename(registry))

        if join_table not in registry.declarativebase.metadata.tables:
            remote_columns, remote_fk = self.get_m2m_columns(
                registry, remote_columns, self.m2m_remote_columns)
            local_columns, local_fk = self.get_m2m_columns(
                registry, local_columns, self.m2m_local_columns)

            Table(join_table, registry.declarativebase.metadata, *(
                local_columns + remote_columns + [local_fk, remote_fk]))

        self.kwargs['secondary'] = join_table
        # definition of expiration
        if self.kwargs.get('backref'):
            self.init_expire_attributes(registry, namespace, fieldname)
            backref = self.kwargs['backref']
            if isinstance(backref, (tuple, list)):
                backref = backref[0]

            registry.expire_attributes[namespace][fieldname].add(
                ('x2m', fieldname, backref))
            model_name = self.model.model_name
            self.init_expire_attributes(registry, model_name, backref)
            registry.expire_attributes[model_name][backref].add(
                ('x2m', backref, fieldname))

        return super(Many2Many, self).get_sqlalchemy_mapping(
            registry, namespace, fieldname, properties)


class InstrumentedAttribute_O2M(attributes.InstrumentedAttribute):

    def __init__(self, *args, **kwargs):
        self.relationship_field = kwargs.pop('relationship_field')
        super(InstrumentedAttribute_O2M, self).__init__(*args, **kwargs)

    def __set__(self, instance, value):
        super(InstrumentedAttribute_O2M, self).__set__(instance, value)
        if instance:
            for cname, fname in self.relationship_field.link_between_columns:
                if value:
                    setattr(instance, cname, getattr(value, fname))
                else:
                    setattr(instance, cname, value)


class One2Many(RelationShip):
    """ Define a relationship attribute on the model

    ::

        @register(Model)
        class TheModel:

            relationship = One2Many(label="The relationship",
                                    model=Model.RemoteModel,
                                    remote_columns="The remote column",
                                    primaryjoin="Join condition"
                                    many2one="themodel")

    If the primaryjoin is not filled then the join condition is
        "'local table'.'local promary key' == 'remote table'.'remote colum'"

    :param model: the remote model
    :param remote_columns: the column name on the remote model
    :param primaryjoin: the join condition between the remote column
    :param many2one: create the many2one link with this one2many
    """

    InstrumentedAttribute = InstrumentedAttribute_O2M

    def __init__(self, **kwargs):
        super(One2Many, self).__init__(**kwargs)

        self.remote_columns = None
        if 'remote_columns' in kwargs:
            remote_columns = self.kwargs.pop('remote_columns')
            if not isinstance(remote_columns, (list, tuple)):
                remote_columns = [remote_columns]

            self.remote_columns = [ModelAttribute(self.model.model_name, x)
                                   for x in remote_columns]

        if 'many2one' in kwargs:
            self.kwargs['backref'] = self.kwargs.pop('many2one')
            self.kwargs['info']['remote_names'] = self.kwargs['backref']

    def autodoc_get_properties(self):
        res = super(One2Many, self).autodoc_get_properties()
        res['remote_columns'] = self.remote_columns
        return res

    def find_foreign_key(self, registry, properties, tablename):
        """ Return the primary key come from the first step property

        :param registry: the registry which load the relationship
        :param properties: first step properties for the model
        :param tablename: the name of the table for the foreign key
        :rtype: column name of the primary key
        """
        fks = []
        for f, p in properties.items():
            if f == '__tablename__':
                continue

            if not hasattr(p, 'foreign_key'):
                continue

            if p.foreign_key:
                model = p.foreign_key.model_name
                if self.get_tablename(registry, model=model) == tablename:
                    fks.append(f)

        return fks

    def add_expire_attributes(self, registry, namespace, fieldname):
        if self.kwargs.get('backref'):
            backref = self.kwargs['backref']
            if isinstance(backref, (list, tuple)):
                backref = backref[0]

            model_name = self.model.model_name
            for rname in self.remote_columns:
                self.init_expire_attributes(
                    registry, model_name, rname.attribute_name)
                _rname = rname.attribute_name
                registry.expire_attributes[model_name][_rname].add((backref,))
                registry.expire_attributes[model_name][_rname].add(
                    (backref, fieldname))

    def get_sqlalchemy_mapping(self, registry, namespace, fieldname,
                               properties):
        """ Create the relationship

        :param registry: the registry which load the relationship
        :param namespace: the name space of the model
        :param fieldname: fieldname of the relationship
        :param propertie: the properties known
        :rtype: Many2One relationship
        """
        self.model.check_model(registry)
        if not self.remote_columns:
            self.remote_columns = self.model.foreign_keys_for(
                registry, namespace)

        self.kwargs['info']['remote_columns'] = [x.attribute_name
                                                 for x in self.remote_columns]

        self.link_between_columns = [
            (x.attribute_name, x.get_fk_mapper(registry).attribute_name)
            for x in self.remote_columns]

        if 'primaryjoin' not in self.kwargs:
            pjs = []
            for cname in self.remote_columns:
                pjs.append("%s == %s" % (
                    cname.get_fk_remote(registry),
                    cname.get_fk_name(registry)))

            # This must be a python and not a SQL AND cause of eval do
            # on the primaryjoin
            self.kwargs['primaryjoin'] = ' and '.join(pjs)

        self.add_expire_attributes(registry, namespace, fieldname)
        return super(One2Many, self).get_sqlalchemy_mapping(
            registry, namespace, fieldname, properties)

    def define_backref_properties(self, registry, namespace, properties):
        """ Add option in the backref if both model and remote model are the
        same, it is for the One2Many on the same model

        :param registry: the registry which load the relationship
        :param namespace: the name space of the model
        :param propertie: the properties known
        """
        if namespace == self.model.model_name:
            pks = ModelRepr(namespace).primary_keys(registry)
            self.backref_properties.update({'remote_side': [
                properties[anyblok_column_prefix + pk.attribute_name]
                for pk in pks]})

    def get_relationship_cls(self):
        self.kwargs['relationship_field'] = self
        return RelationshipProperty



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from anyblok.column import Integer, DateTime, Json, String, Boolean, Text
from threading import Thread
from datetime import datetime
from time import sleep
from sqlalchemy import or_
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm.exc import NoResultFound
from ..exceptions import CronWorkerException
from logging import getLogger
logger = getLogger(__name__)


register = Declarations.register
System = Declarations.Model.System


@register(System)
class Cron:
    started = True

    @classmethod
    def add_worker_for(cls, record):
        worker = cls.registry.System.Cron.Worker(record)
        worker.start()
        return worker

    @classmethod
    def close_worker_with_success(cls, worker):
        worker.job.update(done_at=datetime.now())

    @classmethod
    def close_worker_on_error(cls, worker, error):
        worker.job.update(error=error)

    @classmethod
    def lock_one_job(cls):
        Job = cls.registry.System.Cron.Job
        query = Job.query().filter(Job.done_at.is_(None))
        query = query.filter(or_(Job.available_at.is_(None),
                                 Job.available_at <= datetime.now()))
        nb_available_line = query.count()
        if not nb_available_line:
            return None

        query = query.limit(1)
        for offset in range(nb_available_line):
            try:
                job = query.offset(offset).with_for_update(nowait=True).one()
                return job
            except (OperationalError, NoResultFound):
                cls.registry.rollback()

        return None

    @classmethod
    def run(cls, sleep_time=60, timeout=None):
        sleep_time = 60
        while cls.started:
            job = cls.lock_one_job()
            if job is not None:
                cls.registry.System.Cache.clear_invalidate_cache()
                worker = cls.add_worker_for(job)
                worker.join(timeout)
                error = worker.get_error()
                if error:
                    cls.close_worker_on_error(worker, error)
                else:
                    cls.close_worker_with_success(worker)

                cls.registry.commit()
            else:
                sleep(sleep_time)


@register(System.Cron)
class Job:

    id = Integer(primary_key=True)
    available_at = DateTime()
    done_at = DateTime()
    model = String(nullable=False, foreign_key=System.Model.use('name'))
    method = String(nullable=False)
    is_a_class_method = Boolean(default=True)
    params = Json()
    error = Text()

    def __repr__(self):
        msg = "<Cron job %s.%s" % (self.model, self.method)
        if self.is_a_class_method:
            msg += "(CM)"

        if self.available_at:
            msg += " available at %s" % self.available_at

        msg += " with params %r>" % (self.params)
        return msg


@register(System.Cron)
class Worker(Thread):

    def __init__(self, job):
        super(Worker, self).__init__()
        self.job = job
        self.error = None

    def get_model_record(self):
        params = self.job.params
        if not params or 'primary_keys' not in params:
            raise CronWorkerException("%r : No primary key found in params" %
                                      self.job)
        obj = self.registry.get(self.job.model).from_primary_keys(
            **params['primary_keys'])

        if obj is None:
            raise CronWorkerException("%r : Wrong primary key %r" % (
                self.job, params['primary_keys']))

        return obj

    def get_args_and_kwargs(self):
        params = self.job.params
        args = ()
        kwargs = {}
        if params and 'args' in params:
            args = params['args']

        if isinstance(args, list):
            args = tuple(args)
        elif not isinstance(args, tuple):
            args = (args,)

        if params and 'kwargs' in params:
            kwargs = params['kwargs']

        if not isinstance(kwargs, dict):
            raise CronWorkerException("%r : Kwargs %r must be a dict" % (
                self.job, kwargs))

        return args, kwargs

    def call_method(self):
        if self.job.is_a_class_method:
            record = self.registry.get(self.job.model)
        else:
            record = self.get_model_record()

        func = getattr(record, self.job.method, None)
        if func is None:
            raise CronWorkerException("%r : Inexisting method" % self.job)

        args, kwargs = self.get_args_and_kwargs()
        return func(*args, **kwargs)

    def get_error(self):
        return self.error

    def run(self):
        logger.info("Start worker for %r", self.job)
        try:
            self.call_method()
            self.registry.commit()
        except Exception as e:
            logger.error('Error during execution of %r : %r', self.job, e)
            self.error(str(e))
            self.registry.rollback()
        else:
            logger.info("Worker for %r finish with success", self.job)
        finally:
            self.registry.session.close()



# -*- coding: utf-8 -*-
from sqlalchemy.exc import IntegrityError
from alembic.migration import MigrationContext
from alembic.autogenerate import compare_metadata
from alembic.operations import Operations
from sqlalchemy import schema
from contextlib import contextmanager
from sqlalchemy import func, select, update, join, alias, and_
from anyblok.config import Configuration
import re
from logging import getLogger

logger = getLogger(__name__)


@contextmanager
def cnx(migration):
    """ Context manager used by migration to get the connection """
    try:
        yield migration.conn
    except MigrationException:
        raise
    except Exception:
        migration.conn.execute("rollback")
        raise


class MigrationException(AttributeError):
    """ Simple Exception class for Migration """


class MigrationReport:
    """ Change report

    Get a new report::

        report = MigrationReport(migrationinstance, change_detected)

    """
    def raise_if_withoutautomigration(self):
        if self.migration.withoutautomigration:
            raise MigrationException("The metadata and the base structue are "
                                     "different, or this difference is "
                                     "forbidden in 'no auto migration' mode")

    def init_add_column(self, diff):
        self.raise_if_withoutautomigration()
        _, _, table, column = diff
        self.log_names.append('Add %s.%s' % (table, column.name))

    def check_if_table_and_columns_exist(self, table, *columns):
        try:  # check if the table and the column exist
            table = self.migration.table(table)
            for column in columns:
                table.column(column)
            return True
        except:
            return False

    def can_remove_constraints(self, name):
        unique = "anyblok_uq_(?P<table>\w+)__(?P<columns>\w+)"
        check = "anyblok_ck_(?P<table>\w+)_(?P<constraint>\w+)"

        m = re.search(unique, name)
        if m and self.check_if_table_and_columns_exist(m.group('table'),
                                                       m.group('columns')):
            return True

        m = re.search(check, name)
        if m and self.check_if_table_and_columns_exist(m.group('table')):
            return True

        if self.migration.reinit_constraints:
            return True

        if self.migration.reinit_all:
            return True

        return False

    def can_remove_fk_constraints(self, name):
        fk = ("anyblok_fk_(?P<table>\w+)__(?P<columns>\w+)_on_"
              "(?P<referred_table>\w+)__(?P<referred_columns>\w+)")
        m = re.search(fk, name)
        if m is not None:
            local = self.check_if_table_and_columns_exist(
                m.group('table'), m.group('columns'))
            referred = self.check_if_table_and_columns_exist(
                m.group('referred_table'), m.group('referred_columns'))
            if local and referred:
                return True

        if self.migration.reinit_constraints:
            return True

        if self.migration.reinit_all:
            return True

        return False

    def init_remove_constraint(self, diff):
        _, constraint = diff
        self.log_names.append('Drop constraint %s on %s' % (
            constraint.name, constraint.table))

        if self.can_remove_constraints(constraint.name):
            self.raise_if_withoutautomigration()
        else:
            return True

    def can_remove_index(self, name):
        key = "anyblok_ix_(?P<table>\w+)__(?P<columns>\w+)"
        m = re.search(key, name)
        if m and self.check_if_table_and_columns_exist(m.group('table'),
                                                       m.group('columns')):
            return True

        if self.migration.reinit_indexes:
            return True

        if self.migration.reinit_all:
            return True

        return False

    def init_remove_index(self, diff):
        _, index = diff
        self.log_names.append('Drop index %s on %s' % (index.name,
                                                       index.table))
        if self.can_remove_index(index.name):
            self.raise_if_withoutautomigration()
        else:
            return True

    def init_add_fk(self, diff):
        self.raise_if_withoutautomigration()
        _, fk = diff
        from_ = []
        to_ = []
        for column in fk.columns:
            for fk_ in column.foreign_keys:
                from_.append('%s.%s' % (fk.table.name, column.name))
                to_.append(fk_.target_fullname)

        self.log_names.append('Add Foreign keys on (%s) => (%s)' % (
            ', '.join(from_), ', '.join(to_)))

    def init_remove_fk(self, diff):
        _, fk = diff
        for column in fk.columns:
            for fk_ in column.foreign_keys:
                self.log_names.append('Drop Foreign keys on %s.%s => %s' % (
                    fk.table.name, column.name, fk_.target_fullname))

        if not self.can_remove_fk_constraints(fk.name):
            return True

        self.raise_if_withoutautomigration()

    def init_add_constraint(self, diff):
        self.raise_if_withoutautomigration()
        _, constraint = diff
        columns = [x.name for x in constraint.columns]
        self.log_names.append('Add unique constraint on %s (%s)' % (
            constraint.table.name, ', '.join(columns)))

    def can_remove_column(self):
        if self.migration.reinit_columns:
            return True

        if self.migration.reinit_all:
            return True

        return False

    def init_remove_column(self, diff):
        column = diff[3]
        msg = "Drop Column %s.%s" % (column.table.name,
                                     column.name)

        if self.can_remove_column():
            self.log_names.append(msg)
            self.raise_if_withoutautomigration()
            return False

        fk_removed = []
        for fk in column.foreign_keys:
            if not self.can_remove_fk_constraints(fk.name):
                # only if fk is not removable. FK can come from
                # * DBA manager, it is the only raison to destroy it
                # * alembic, some constrainte change name during the remove
                if fk.name not in fk_removed:
                    self.actions.append(('remove_fk', fk.constraint))
                    fk_removed.append(fk.name)

        if column.nullable is False:
            self.raise_if_withoutautomigration()
            msg += " (not null)"
            self.log_names.append(msg)
            self.actions.append(
                ('modify_nullable', None, column.table.name,
                 column.name, {}, False, True))
            return True

        self.log_names.append(msg)
        return True

    def can_remove_table(self):
        if self.migration.reinit_tables:
            return True

        if self.migration.reinit_all:
            return True

        return False

    def init_remove_table(self, diff):
        self.log_names.append("Drop Table %s" % diff[1].name)
        if self.can_remove_table():
            self.raise_if_withoutautomigration()
        else:
            return True

    def __init__(self, migration, diffs):
        """ Initializer

        :param migration: migration instance
        :param diffs: diff between the metadata and the database, come from
            change detection of alembic
        """
        self.migration = migration
        self.logs = []
        self.actions = []
        self.diffs = diffs
        self.log_names = []
        mappers = {
            'add_column': self.init_add_column,
            'remove_constraint': self.init_remove_constraint,
            'remove_index': self.init_remove_index,
            'add_fk': self.init_add_fk,
            'remove_fk': self.init_remove_fk,
            'add_constraint': self.init_add_constraint,
            'remove_column': self.init_remove_column,
            'remove_table': self.init_remove_table,
        }
        for diff in diffs:
            if isinstance(diff, list):
                self.raise_if_withoutautomigration()
                for change in diff:
                    _, _, table, column, _, _, _ = change
                    self.log_names.append('Alter %s.%s' % (table, column))
                    self.actions.append(change)
            else:
                fnct = mappers.get(diff[0])
                if fnct:
                    if fnct(diff):
                        continue
                else:
                    print(diff)

                self.actions.append(diff)

        for log_name in self.log_names:
            if log_name and not self.log_has(log_name):
                self.logs.append(log_name)

    def log_has(self, log):
        """ return True id the log is present

        .. warning:: this method is only used for the unittest

        :param log: log sentence expected
        """
        return log in self.logs

    def apply_change_add_column(self, action):
        _, _, table, column = action
        self.migration.table(table).column().add(column)

    def apply_change_modify_nullable(self, action):
        _, _, table, column, kwargs, oldvalue, newvalue = action
        self.migration.table(table).column(column).alter(
            nullable=newvalue, existing_nullable=oldvalue, **kwargs)

    def apply_change_modify_type(self, action):
        _, _, table, column, kwargs, oldvalue, newvalue = action
        self.migration.table(table).column(column).alter(
            type_=newvalue, existing_type=oldvalue, **kwargs)

    def apply_change_modify_default(self, action):
        _, _, table, column, kwargs, oldvalue, newvalue = action
        self.migration.table(table).column(column).alter(
            server_default=newvalue, existing_server_default=oldvalue,
            **kwargs)

    def apply_change_remove_constraint(self, action):
        # FIXME, test for check, foreignkey
        _, constraint = action
        if constraint.__class__ is schema.UniqueConstraint:
            table = self.migration.table(constraint.table)
            table.unique(name=constraint.name).drop()

    def apply_change_remove_index(self, action):
        _, index = action
        if not index.unique:
            table = self.migration.table(index.table.name)
            table.index(name=index.name).drop()

    def apply_change_add_fk(self, action):
        _, fk = action
        t = self.migration.table(fk.table.name)
        from_ = []
        to_ = []
        for column in fk.columns:
            for fk_ in column.foreign_keys:
                from_.append(column.name)
                to_.append(fk_.column)

        name = 'fk_%s' % '_'.join(from_)
        t.foreign_key(name).add(from_, to_)

    def apply_change_remove_fk(self, action):
        _, fk = action
        t = self.migration.table(fk.table.name)
        t.foreign_key(fk.name).drop()

    def apply_change_add_constraint(self, action):
        _, constraint = action
        table = self.migration.table(constraint.table.name)
        # FIXME, test for check, foreignkey
        table.unique(name=constraint.name).add(*constraint.columns)

    def apply_remove_table(self, action):
        self.migration.table(action[1].name).drop()

    def apply_remove_column(self, action):
        self.migration.table(action[2]).column(action[3].name).drop()

    def apply_change(self):
        """ Apply the migration

        this method parses the detected change and calls the Migration
        system to apply the change with the api of Declarations
        """
        for log in self.logs:
            logger.info(log)

        mappers = {
            'add_column': self.apply_change_add_column,
            'modify_nullable': self.apply_change_modify_nullable,
            'modify_type': self.apply_change_modify_type,
            'modify_default': self.apply_change_modify_default,
            'add_fk': self.apply_change_add_fk,
            'add_constraint': self.apply_change_add_constraint,
            'remove_constraint': self.apply_change_remove_constraint,
            'remove_index': self.apply_change_remove_index,
            'remove_fk': self.apply_change_remove_fk,
            'remove_table': self.apply_remove_table,
            'remove_column': self.apply_remove_column,
        }

        for action in self.actions:
            fnct = mappers.get(action[0])
            if fnct:
                fnct(action)


class MigrationConstraintForeignKey:
    """ Used to apply a migration on a foreign key

    You can add::

        table.column('my column').foreign_key().add(Blok.name)

    Or drop::

        table.column('my column').foreign_key().drop()

    """
    def __init__(self, table, name):
        self.table = table
        self.name = name

    def add(self, local_columns, remote_columns, **kwargs):
        """ Add a new foreign key

        :param remote_field: The column of the remote model
        :rtype: MigrationConstraintForeignKey instance
        """
        remote_columns = [
            x.property.columns[0] if hasattr(x, 'property') else x
            for x in remote_columns]

        remote_table = set(x.table.name for x in remote_columns)
        if len(remote_table) != 1:
            raise MigrationException("Remote column must have the same table "
                                     "(%s)" % ', '.join(remote_table))

        remote_table = remote_table.pop()
        remote_columns = [x.name for x in remote_columns]
        self.table.migration.operation.create_foreign_key(
            self.name, self.table.name, remote_table,
            local_columns, remote_columns, **kwargs)
        return self

    def drop(self):
        """ Drop the foreign key """
        self.table.migration.operation.drop_constraint(
            self.name, self.table.name, type_='foreignkey')
        return self


class MigrationColumn:
    """ get or add a column

    Add a new column::

        table.column().add(Sqlachemy column)

    Get a column::

        c = table.column('My column name')

    Alter the column::

        c.alter(new_column_name='Another column name')

    Drop the column::

        c.drop()
    """
    def __init__(self, table, name):
        self.table = table
        self.name = name
        self.info = {}

        if name is not None:
            op = self.table.migration.operation
            with cnx(self.table.migration) as conn:
                columns = op.impl.dialect.get_columns(
                    conn, self.table.name)

            for c in columns:
                if c['name'] == name:
                    self.info.update(c)

            if not self.info:
                raise MigrationException(
                    "No column %r found on %r" % (name, self.table.name))

    def apply_default_value(self, column):
        if column.default:
            execute = self.table.migration.conn.execute
            val = column.default.arg
            table = self.table.migration.metadata.tables[self.table.name]
            table.append_column(column)
            cname = getattr(table.c, column.name)
            if column.default.is_callable:
                table2 = alias(select([table]).limit(1).where(cname.is_(None)))
                Table = self.table.migration.metadata.tables['system_model']
                Column = self.table.migration.metadata.tables['system_column']
                j1 = join(Table, Column, Table.c.name == Column.c.model)
                query = select([func.count()]).select_from(table)
                nb_row = self.table.migration.conn.execute(query).fetchone()[0]
                query = select([Column.c.name]).select_from(j1)
                query = query.where(Column.c.primary_key.is_(True))
                query = query.where(Table.c.table == self.table.name)
                columns = [x[0] for x in execute(query).fetchall()]
                where = and_(*[getattr(table.c, x) == getattr(table2.c, x)
                               for x in columns])
                for offset in range(nb_row):
                    # call for each row because the default value
                    # could be a sequence or depend of other field
                    query = update(table).where(where).values(
                        {cname: val(None)})
                    execute(query)

            else:
                query = update(table).where(cname.is_(None)).values(
                    {cname: val})
                execute(query)

    def add(self, column):
        """ Add a new column

        The column is added in two phases, the last phase is only for the
        the nullable, if nullable can not be applied, a warning is logged

        :param column: sqlalchemy column
        :rtype: MigrationColumn instance
        """
        nullable = column.nullable
        if not nullable:
            column.nullable = True

        self.table.migration.operation.impl.add_column(self.table.name, column)
        self.apply_default_value(column)

        if not nullable:
            c = MigrationColumn(self.table, column.name)
            c.alter(existing_type=column.type,
                    existing_server_default=column.server_default,
                    existing_autoincrement=column.autoincrement,
                    existing_nullable=True, nullable=False)

        t = self.table.migration.metadata.tables[self.table.name]
        for constraint in t.constraints:
            if not isinstance(constraint, schema.PrimaryKeyConstraint):
                if not isinstance(constraint, schema.ForeignKeyConstraint):
                    self.table.migration.operation.impl.add_constraint(
                        constraint)

        # TODO get the default value of the column and apply it on null value

        return MigrationColumn(self.table, column.name)

    def alter(self, **kwargs):
        """ Alter an existing column

        Alter the column in two phases, because the nullable column has not
        locked the migration

        .. warning::
            See Alembic alter_column, the existing_* param are used for some
            dialect like mysql, is importante to filled them for these dialect

        :param new_column_name: New name for the column
        :param type_: New sqlalchemy type
        :param existing_type: Old sqlalchemy type
        :param server_default: The default value in database server
        :param existing_server_default: Old default value
        :param nullable: New nullable value
        :param existing_nullable: Old nullable value
        :param autoincrement: New auto increment use for Integer whith primary
            key only
        :param existing_autoincrement: Old auto increment
        :rtype: MigrationColumn instance
        """
        vals = {}
        name = self.name

        for k in ('existing_type', 'existing_server_default',
                  'existing_nullable', 'existing_autoincrement',
                  'autoincrement', 'server_default', 'type_'):
            if k in kwargs:
                vals[k] = kwargs[k]

        if 'name' in kwargs:
            vals['new_column_name'] = kwargs['name']
            name = kwargs['name']

        if vals:
            self.table.migration.operation.alter_column(
                self.table.name, self.name, **vals)

        if 'nullable' in kwargs:
            nullable = kwargs['nullable']
            savepoint = '%s_not_null' % name
            try:
                self.table.migration.savepoint(savepoint)
                self.table.migration.operation.alter_column(
                    self.table.name, self.name, nullable=nullable, **vals)
            except IntegrityError as e:
                self.table.migration.rollback_savepoint(savepoint)
                logger.warn(str(e))

        return MigrationColumn(self.table, name)

    def drop(self):
        """ Drop the column """
        self.table.migration.operation.drop_column(self.table.name, self.name)

    def nullable(self):
        """ Use for unittest return if the column is nullable """
        return self.info['nullable']

    def type(self):
        """ Use for unittest: return the column type """
        return self.info['type']

    def server_default(self):
        """ Use for unittest: return the default database value """
        return self.info['default']


class MigrationConstraintCheck:
    """ Used for the Check constraint

    Add a new constraint::

        table('My table name').check().add('check_my_column', 'mycolumn > 5')

    Get and drop the constraint::

        table('My table name').check('check_my_column').drop()

    """
    def __init__(self, table, name):
        self.table = table
        self.name = name
        # TODO dialect not have method to check if constraint exist

    def add(self, condition):
        """ Add the constraint

        :param condition: constraint to apply
        :rtype: MigrationConstraintCheck instance
        """
        self.table.migration.operation.create_check_constraint(
            self.name, self.table.name, condition)

        return self

    def drop(self):
        """ Drop the constraint """
        self.table.migration.operation.drop_constraint(
            self.name, self.table.name, type_='check')


class MigrationConstraintUnique:
    """ Used for the Unique constraint

    Add a new constraint::

        table('My table name').unique('constraint name').add('col1', 'col2')

    Get and drop the constraint::

        table('My table name').unique('constraint name').drop()

    Let AnyBlok to define the name of the constraint::

        table('My table name').unique(None).add('col1', 'col2')


    """
    def __init__(self, table, name):
        self.table = table
        self.name = name

    def add(self, *columns):
        """ Add the constraint

        :param \*column: list of column name
        :rtype: MigrationConstraintUnique instance
        :exception: MigrationException
        """
        if not columns:
            raise MigrationException("""To add an unique constraint you """
                                     """must define one or more columns""")

        columns_name = [x.name for x in columns]
        savepoint = 'add_unique_constraint_%s' % (self.name or '')
        try:
            self.table.migration.savepoint(savepoint)
            self.table.migration.operation.create_unique_constraint(
                self.name, self.table.name, columns_name)
        except IntegrityError as e:
            self.table.migration.rollback_savepoint(savepoint)
            logger.warn("Error during the add of new unique constraint %r on "
                        "table %r and columns %r : %r " % (self.name,
                                                           self.table.name,
                                                           columns_name,
                                                           str(e)))

        return self

    def drop(self):
        """ Drop the constraint """
        self.table.migration.operation.drop_constraint(
            self.name, self.table.name, type_='unique')


class MigrationConstraintPrimaryKey:
    """ Used for the primary key constraint

    Add a new constraint::

        table('My table name').primarykey().add('col1', 'col2')

    Get and drop the constraint::

        table('My table name').primarykey('col1', 'col2').drop()
    """

    def __init__(self, table):
        self.table = table
        self.name = self.format_name()

    def format_name(self, *columns):
        return 'anyblok_pk_%s' % self.table.name

    def add(self, *columns):
        """ Add the constraint

        :param \*column: list of column name
        :rtype: MigrationConstraintPrimaryKey instance
        :exception: MigrationException
        """
        if not columns:
            raise MigrationException("""To add a primary key constraint """
                                     """you must define one or more columns""")

        columns_name = [x.name for x in columns]
        self.table.migration.operation.create_primary_key(
            self.name, self.table.name, columns_name)
        return self

    def drop(self):
        """ Drop the constraint """
        self.table.migration.operation.drop_constraint(
            self.name, self.table.name, type_='primary')
        return self


class MigrationIndex:
    """ Used for the index constraint

    Add a new constraint::

        table('My table name').index().add('col1', 'col2')

    Get and drop the constraint::

        table('My table name').index('col1', 'col2').drop()

    """

    def __init__(self, table, *columns, **kwargs):
        self.table = table
        if 'name' in kwargs:
            self.name = kwargs['name']
        else:
            self.name = self.format_name(*columns)
        self.exist = False

        if self.name is not None:
            op = self.table.migration.operation
            with cnx(self.table.migration) as conn:
                indexes = op.impl.dialect.get_indexes(
                    conn, self.table.name, None)

            for i in indexes:
                if i['name'] == self.name:
                    self.exist = True

            if not self.exist:
                raise MigrationException(
                    "No index %r found on %r" % (self.name, self.table.name))

    def format_name(self, *columns):
        if columns:
            cols = [x.name for x in columns]
            cols.sort()
            cols = '_'.join(cols)
            return 'idx_%s_on_%s' % (cols, self.table.name)

        return None

    def add(self, *columns):
        """ Add the constraint

        :param \*column: list of column name
        :rtype: MigrationIndex instance
        :exception: MigrationException
        """
        if not columns:
            raise MigrationException(
                "To add an index you must define one or more columns")

        index_name = self.format_name(*columns)
        columns_name = [x.name for x in columns]
        self.table.migration.operation.create_index(
            index_name, self.table.name, columns_name)

        return MigrationIndex(self.table, *columns)

    def drop(self):
        """ Drop the constraint """
        self.table.migration.operation.drop_index(
            self.name, table_name=self.table.name)


class MigrationTable:
    """ Use to manipulate tables

    Add a table::

        table().add('New table')

    Get an existing table::

        t = table('My table name')

    Alter the table::

        t.alter(name='Another table name')

    Drop the table::

        t.drop()
    """

    def __init__(self, migration, name):
        self.name = name
        self.migration = migration

        if name is not None:
            with cnx(self.migration) as conn:
                if not self.migration.operation.impl.dialect.has_table(conn,
                                                                       name):
                    raise MigrationException("No table %r found" % name)

    def add(self, name):
        """ Add a new table

        :param name: name of the table
        :rtype: MigrationTable instance
        """
        self.migration.operation.create_table(name)
        return MigrationTable(self.migration, name)

    def column(self, name=None):
        """ Get Column

        :param name: Column name
        :rtype: MigrationColumn instance
        """
        return MigrationColumn(self, name)

    def drop(self):
        """ Drop the table """
        self.migration.operation.drop_table(self.name)

    def index(self, *columns, **kwargs):
        """ Get index

        :param \*columns: List of the column's name
        :rtype: MigrationIndex instance
        """
        return MigrationIndex(self, *columns, **kwargs)

    def unique(self, name):
        """ Get unique

        :param \*columns: List of the column's name
        :rtype: MigrationConstraintUnique instance
        """
        return MigrationConstraintUnique(self, name)

    def check(self, name=None):
        """ Get check

        :param \*columns: List of the column's name
        :rtype: MigrationConstraintCheck instance
        """
        return MigrationConstraintCheck(self, name)

    def primarykey(self):
        """ Get primary key

        :param \*columns: List of the column's name
        :rtype: MigrationConstraintPrimaryKey instance
        """
        return MigrationConstraintPrimaryKey(self)

    def alter(self, **kwargs):
        """ Atler the current table

        :param name: New table name
        :rtype: MigrationTable instance
        :exception: MigrationException
        """
        if 'name' not in kwargs:
            raise MigrationException("Table can only alter name")

        name = kwargs['name']
        self.migration.operation.rename_table(self.name, name)
        return MigrationTable(self.migration, name)

    def foreign_key(self, name):
        """ Get a foreign key

        :rtype: MigrationConstraintForeignKey instance
        """
        return MigrationConstraintForeignKey(self, name)


class Migration:
    """ Migration Main entry

    This class allows to manipulate all the migration class::

        migration = Migration(Session(), Base.Metadata)
        t = migration.table('My table name')
        c = t.column('My column name from t')
    """

    def __init__(self, registry):
        self.withoutautomigration = registry.withoutautomigration
        self.conn = registry.session.connection()
        self.metadata = registry.declarativebase.metadata

        opts = {
            'compare_server_default': True,
            'render_item': self.render_item,
            'compare_type': self.compare_type,
        }
        self.context = MigrationContext.configure(self.conn, opts=opts)
        self.operation = Operations(self.context)
        self.reinit_all = Configuration.get('reinit_all', False)
        self.reinit_tables = Configuration.get('reinit_tables', False)
        self.reinit_columns = Configuration.get('reinit_columns', False)
        self.reinit_indexes = Configuration.get('reinit_indexes', False)
        self.reinit_constraints = Configuration.get(
            'reinit_constraints', False)

    def table(self, name=None):
        """ Get a table

        :rtype: MigrationTable instance
        """
        return MigrationTable(self, name)

    def auto_upgrade_database(self):
        """ Upgrade the database automaticly """
        report = self.detect_changed()
        report.apply_change()

    def detect_changed(self):
        """ Detect the difference between the metadata and the database

        :rtype: MigrationReport instance
        """
        diff = compare_metadata(self.context, self.metadata)
        return MigrationReport(self, diff)

    def savepoint(self, name=None):
        """ Add a savepoint

        :param name: name of the save point
        :rtype: return the name of the save point
        """
        return self.conn._savepoint_impl(name=name)

    def rollback_savepoint(self, name):
        """ Rollback to the savepoint

        :param name: name of the savepoint
        """
        self.conn._rollback_to_savepoint_impl(name, None)

    def release_savepoint(self, name):
        """ Release the save point

        :param name: name of the savepoint
        """
        self.conn._release_savepoint_impl(name, None)

    def render_item(self, type_, obj, autogen_context):
        logger.debug("%r, %r, %r" % (type_, obj, autogen_context))
        return False

    def compare_type(self, context, inspected_column,
                     metadata_column, inspected_type, metadata_type):
        if hasattr(metadata_type, 'compare_type'):
            return metadata_type.compare_type(inspected_type)

        return None



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import BlokTestCase


class TestCoreSqlBase(BlokTestCase):

    def test_insert(self):
        Blok = self.registry.System.Blok
        Blok.insert(name='OneBlok', state='undefined', version='0.0.0')
        blok = Blok.query().filter(Blok.name == 'OneBlok').first()
        self.assertEqual(blok.state, 'undefined')

    def test_multi_insert(self):
        Blok = self.registry.System.Blok
        Blok.multi_insert(
            dict(name='OneBlok', state='undefined', version='0.0.0'),
            dict(name='TwoBlok', state='undefined', version='0.0.0'),
            dict(name='ThreeBlok', state='undefined', version='0.0.0'))
        states = Blok.query('state').filter(Blok.name.in_(['OneBlok',
                                                           'TwoBlok',
                                                           'ThreeBlok'])).all()
        states = [x[0] for x in states]
        self.assertEqual(states, ['undefined', 'undefined', 'undefined'])

    def test_from_primary_key(self):
        Model = self.registry.System.Model
        model = Model.query().first()
        model2 = Model.from_primary_keys(name=model.name)
        self.assertEqual(model, model2)

    def test_from_primary_keys(self):
        Column = self.registry.System.Column
        column = Column.query().first()
        column2 = Column.from_primary_keys(model=column.model,
                                           name=column.name)
        self.assertEqual(column, column2)

    def test_get_primary_key(self):
        self.assertEqual(self.registry.System.Model.get_primary_keys(),
                         ['name'])

    def test_get_primary_keys(self):
        pks = self.registry.System.Column.get_primary_keys()
        model_in_pks = 'model' in pks
        name_in_pks = 'name' in pks
        self.assertEqual(model_in_pks, True)
        self.assertEqual(name_in_pks, True)

    def test_to_primary_key(self):
        model = self.registry.System.Model.query().first()
        self.assertEqual(model.to_primary_keys(), dict(name=model.name))

    def test_to_primary_keys(self):
        column = self.registry.System.Column.query().first()
        self.assertEqual(column.to_primary_keys(),
                         {'model': column.model, 'name': column.name})

    def test_fields_description(self):
        Model = self.registry.System.Model
        self.maxDiff = None
        res = {'description': {'id': 'description',
                               'label': 'Description',
                               'model': None,
                               'nullable': True,
                               'primary_key': False,
                               'type': 'Function'},
               'is_sql_model': {'id': 'is_sql_model',
                                'label': 'Is a SQL model',
                                'model': None,
                                'nullable': True,
                                'primary_key': False,
                                'type': 'Boolean'},
               'name': {'id': 'name',
                        'label': 'Name',
                        'model': None,
                        'nullable': False,
                        'primary_key': True,
                        'type': 'String'},
               'table': {'id': 'table',
                         'label': 'Table',
                         'model': None,
                         'nullable': True,
                         'primary_key': False,
                         'type': 'String'}}
        self.assertEqual(Model.fields_description(), res)

    def test_fields_description_limited_field(self):
        Model = self.registry.System.Model
        self.maxDiff = None
        res = {'table': {'id': 'table',
                         'label': 'Table',
                         'model': None,
                         'nullable': True,
                         'primary_key': False,
                         'type': 'String'}}
        self.assertEqual(Model.fields_description(fields=['table']), res)

    def test_fields_description_cache(self):
        Model = self.registry.System.Model
        Column = self.registry.System.Column
        self.maxDiff = None
        res = {'table': {'id': 'table',
                         'label': 'Table',
                         'model': None,
                         'nullable': True,
                         'primary_key': False,
                         'type': 'String'}}
        self.assertEqual(Model.fields_description(fields=['table']), res)
        column = Column.from_primary_keys(model='Model.System.Model',
                                          name='table')
        column.label = 'Test'
        self.assertEqual(Model.fields_description(fields=['table']), res)
        Model.fire('Update Model', 'Model.System.Model')
        self.assertNotEqual(Model.fields_description(fields=['table']), res)

    def test_to_dict(self):
        M = self.registry.System.Model
        model = M.query().first()
        self.assertEqual(model.to_dict(), {
            'name': model.name,
            'table': model.table,
            'is_sql_model': model.is_sql_model,
            'description': model.description,
        })

    def test_to_dict_on_some_columns(self):
        M = self.registry.System.Model
        model = M.query().first()
        self.assertEqual(model.to_dict('name', 'table'), {
            'name': model.name,
            'table': model.table,
        })



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import TestCase
from anyblok._graphviz import ModelSchema, SQLSchema


class TestSQLSchema(TestCase):

    def test_one_table(self):
        dot = SQLSchema('My table model')
        dot.add_table('My table')
        dot.render()

    def test_one_label(self):
        dot = SQLSchema('My table model')
        dot.add_table('My table')
        dot.render()

    def test_one_table_with_column(self):
        dot = SQLSchema('My table model')
        table = dot.add_table('My table')
        table.add_column('My column', 'Integer', primary_key=True)
        dot.render()

    def test_one_table_with_foreign_key(self):
        dot = SQLSchema('My table model')
        table = dot.add_table('My table')
        table.add_column('My column', 'Integer', primary_key=True)
        table2 = dot.add_table('A 2nd table')
        table2.add_foreign_key('My fk', table, nullable=False)
        dot.render()

    def test_one_table_with_foreign_key_on_label(self):
        dot = SQLSchema('My table model')
        table = dot.add_label('My table')
        table.add_column('My column', 'Integer', primary_key=True)
        table2 = dot.add_table('A 2nd table')
        table2.add_foreign_key('My fk', table, nullable=False)
        dot.render()

    def test_get_table(self):
        dot = SQLSchema('My UML model')
        table = dot.add_table('My table')
        self.assertEqual(dot.get_table('My table'), table)

    def test_get_table_with_label(self):
        dot = SQLSchema('My UML model')
        table = dot.add_label('My table')
        self.assertEqual(dot.get_table('My table'), table)


class TestModelSchema(TestCase):

    def test_one_model_with_label(self):
        dot = ModelSchema('My UML model')
        dot.add_label('My label')
        dot.render()

    def test_one_model_with_class(self):
        dot = ModelSchema('My UML model')
        dot.add_class('My class')
        dot.render()

    def test_one_model_with_class_which_extend_label(self):
        dot = ModelSchema('My UML model')
        label = dot.add_label('My label')
        cls = dot.add_class('My class')
        cls.add_method('One method')
        cls.extend(label)
        dot.render()

    def test_one_model_with_class_which_associate_label(self):
        dot = ModelSchema('My UML model')
        label = dot.add_label('My label')
        cls = dot.add_class('My class')
        cls.associate(label)
        cls.associate(cls, label_from='test')
        dot.render()

    def test_one_model_with_class_which_agregate_label(self):
        dot = ModelSchema('My UML model')
        label = dot.add_label('My label')
        cls = dot.add_class('My class')
        cls.agregate(label, label_to='test', multiplicity_to='test')
        dot.render()

    def test_one_model_with_class_which_strong_agregate_label(self):
        dot = ModelSchema('My UML model')
        label = dot.add_label('My label')
        cls = dot.add_class('My class')
        cls.strong_agregate(label, multiplicity_from='test')
        dot.render()

    def test_get_class(self):
        dot = ModelSchema('My UML model')
        cls = dot.add_class('My class')
        self.assertEqual(dot.get_class('My class'), cls)

    def test_get_class_with_label(self):
        dot = ModelSchema('My UML model')
        cls = dot.add_label('My class')
        self.assertEqual(dot.get_class('My class'), cls)
